/**
 * Copyright (c) 2009 Symantec
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the 'Software'), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *  
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *  
 * THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package com.symantec.premiumServices.asyncThreading {
	
	import com.symantec.premiumServices.asyncThreading.events.AsyncThreadManagerEvent;
	import com.symantec.premiumServices.asyncThreading.handlers.AsyncThreadHandler;
	import com.symantec.premiumServices.asyncThreading.handlers.FPSObserverHandler;
	import com.symantec.premiumServices.asyncThreading.interfaces.*;
	import com.symantec.premiumServices.asyncThreading.namespaces.async_threading;
	import com.symantec.premiumServices.asyncThreading.vo.AsyncThreadMessage;
	
	import flash.display.DisplayObject;
	import flash.events.EventDispatcher;
	
	import mx.logging.Log;
	
	/**
	 * Dispatched when an <code>IAsyncThreadMessage</code> is introduced to the system.
	 * 
	 * @eventType com.symantec.premiumServices.asyncThreading.events.AsyncThreadManagerEvent
	*/
	[Event(name="sendThreadMessage", type="com.symantec.premiumServices.asyncThreading.events.AsyncThreadManagerEvent")]
	/**
	 * this is a singleton manager class that manages the handler classes for <code>FPSObserver</code> 
	 * and <code>IAsyncThreadResponder</code> instances.
	 * 
	 * @author ben rimbey -> ben_rimbey@symantec.com, ben.rimbey@gmail.com 
	 * 
	 */
	public class AsyncThreadingManager extends EventDispatcher {
		
		/**
		 * used to store the singleton instance of this class. 
		 */
		protected static var instance:AsyncThreadingManager;
		
		/**
		 * enabling this value allows the AsyncThreading library to log statusing to the console.
		 */
		public static var verboseLogging:Boolean = false;
		/**
		 * enabling this value allows the AsyncThreading library to distribute the threading 
		 * workload over the entire framerate range for each second. the downside is that since 
		 * the distribution is random, it is harder to control timing between multiple threads. 
		 * the assumption, though, is that no more than one thread will work on the same chunk of 
		 * data at a time, which means that this shouldnt be a problem. to that end, the value 
		 * defaults to true in order to maximize performance.
		 * 
		 * <p>disabling this value will force all threads with the same priority to fire on the 
		 * same frames. for example, all threads with a priority of 8 will fire on the same 8th frame, 
		 * instead of using seeds to fire on different 8th frames.</p> 
		 */
		public static var useDistributedFrameLoading:Boolean = true;
		/**
		 * controls the length of <code>FPSReportQueue</code> instances. 
		 */
		public static var fpsMonitorQueueLength:int = 3;
		
		private var _threadHandler:AsyncThreadHandler;
		private var _observerHandler:FPSObserverHandler;
		
		/**
		 * the maximum amount of time that a thread should be allowed to run. only applies to 
		 * threads implementing the IAsyncThreadScalableResponder interface. 
		 */
		public var maxExecutionTime:int = 70;
		/**
		 * the median amount of time that a thread should be allowed to run. only applies to 
		 * threads implementing the IAsyncThreadScalableResponder interface. 
		 */
		public var medianExecutionTime:int = 50;
		/**
		 * defines what constitutes as a 'low' framerate.  
		 */
		public var lowFrameRateThreshold:int = 12;
		/**
		 * defines what constitutes as a 'moderate' framerate.  
		 */
		public var moderateFrameRateThreshold:int = 18;
		
		/**
		 * Returns the singleton instance of this class.
		 * 
		 * @return AsyncThreadManager
		 * 
		 */
		public static function getInstance():AsyncThreadingManager {
			if (instance == null) { 
				instance = new AsyncThreadingManager(new SingletonEnforcer());
				instance.async_threading::initialize(); 
			}
			return instance;
		}
		
		/**
		 * constructor method. requires that the <code>SingletonEnforcer</code> class be passed 
		 * in, which limits who can instantiate this class.
		 *  
		 * @param key
		 * 
		 */
		public function AsyncThreadingManager(key:SingletonEnforcer):void {}
		
		/**
		 * invoked by the <code>getInstance</code> static method when the singleton instance is first 
		 * instantiated. 
		 * 
		 * <p>here we instantiate new <code>AsyncThreadHandler</code> and <code>FPSObserverHandler</code> 
		 * instances. additionally, we invoke the <code>startup</code> method on the 
		 * <code>FPSObserverHandler</code> instance.
		 * 
		 * @return Boolean 
		 * 
		 */
		async_threading function initialize():Boolean {
			async_threading::logDebug("com.symantec.premiumServices.asyncThreading.AsyncThreadManager", "singleton manager class initialized...");
			
			try {
				_threadHandler = new AsyncThreadHandler();
				_observerHandler = new FPSObserverHandler();
				
				_observerHandler.async_threading::startUp();
			} catch (e:Error) {
				throw e;
				return false;
			}
			
			return true;
		}
		
		/**
		 * used internally by the AsyncThreading framework to log statusing to the console. 
		 * 
		 * <p>logging is only available if the <code>verboseLogging</code> value is set to <code>true</code>.</p>
		 * 
		 * <p>if for some reason we are unable to log to the console, we log the error and return 
		 * <code>false</code>.
		 * 
		 * @param classPath
		 * @param data
		 * @return Boolean 
		 * 
		 */
		async_threading function logDebug(classPath:String, data:*):Boolean {
			try {
				if (AsyncThreadingManager.verboseLogging) {
					Log.getLogger(classPath).debug(data);
				}
			} catch (e:Error) {
				async_threading::logDebug("com.symantec.premiumServices.asyncThreading.AsyncThreadManager", "unable to create log entry :: " + data);
				return false;
			}
			
			return true;	
		}
		
		/**
		 * invokes the <code>killAllThreads</code> method on the <code>AsyncThreadHandler</code> instance. 
		 * 
		 */
		public function killAllThreads():void {
			_threadHandler.async_threading::killAllThreads();
		}
		
		/**
		 * passes the <code>uiRef</code> and <code>onEnterFrameCallBack</code> values to the 
		 * <code>registerUIComponent</code> method on the <code>FPSObserverHandler</code> instance.
		 * 
		 * <p>we return the response of this method invocation, which is an instance of 
		 * <code>IFPSObserver</code>.
		 * 
		 * @param uiRef
		 * @param onEnterFrameCallBack
		 * @return IFPSObserver 
		 * 
		 */
		public function monitorDisplayObject(uiRef:DisplayObject, onEnterFrameCallBack:Function=null):IFPSObserver {
			async_threading::logDebug("com.symantec.premiumServices.asyncThreading.AsyncThreadManager", "monitoring display object :: " + uiRef.name);
			return _observerHandler.async_threading::registerUIComponent(uiRef, onEnterFrameCallBack);
		}
		
		/**
		 * attempts to unregister the DisplayObject that is passed in by passing the <code>uiRef</code> 
		 * parameter to the <code>unregisterUIComponent</code> method on the <code>FPSObserverHandler</code> instance.
		 * 
		 * <p>if for some reason we are unable to perform this action, we log the error and return 
		 * <code>false</code>.
		 * 
		 * @param uiRef
		 * @return Boolean 
		 * 
		 */
		public function stopMonitoringDisplayObject(uiRef:DisplayObject):Boolean {
			async_threading::logDebug("com.symantec.premiumServices.asyncThreading.AsyncThreadManager", "halting subsequent monitoring of display object :: " + uiRef.name);
			try {
				_observerHandler.async_threading::unregisterUIComponent(uiRef.name);
			} catch (e:Error) {
				async_threading::logDebug("com.symantec.premiumServices.asyncThreading.AsyncThreadManager", "unable to stop monitoring display object!! :: " + uiRef.name);
				return false;
			}
			
			return true;
		}
		
		/**
		 * creates an <code>AsyncThreadMessage</code> instance based on the <code>name</code> 
		 * and <code>body</code> parameters and dispatches it to the <code>AsyncThreadHandler</code> 
		 * instance for management between threads.
		 * 
		 * <p>if for some reason we are unable to perform this action, we log the error and return 
		 * <code>false</code>.
		 * 
		 * @param name
		 * @param body
		 * @return Boolean 
		 * 
		 */
		public function sendMessageToThreads(name:String, body:Object):Boolean {
			async_threading::logDebug("com.symantec.premiumServices.asyncThreading.AsyncThreadManager", "sending message to threads, key :: " + name);
			try {
				var messageVO:AsyncThreadMessage = new AsyncThreadMessage();
					messageVO.name = name.toLowerCase();
					messageVO.body = body;
				
				_threadHandler.async_threading::handleInboundMessage(messageVO);
				
				var event:AsyncThreadManagerEvent = new AsyncThreadManagerEvent(AsyncThreadManagerEvent.SEND_THREAD_MESSAGE);
					event.message = messageVO;
					
				super.dispatchEvent(event);
			} catch (e:Error) {
				async_threading::logDebug("com.symantec.premiumServices.asyncThreading.AsyncThreadManager", "unable to send message to threads, name :: " + name);
				return false;
			}
			
			return true;
		}
		
		/**
		 * calls the <code>shutDown</code> methods on both the <code>AsyncThreadHandler</code> and <code>FPSObserverHandler</code> 
		 * instances. if either of them return false, we log to the console and return false.
		 *  
		 * @return Boolean 
		 * 
		 */
		public function shutDown():Boolean {
			if (_threadHandler.async_threading::shutDown() && _observerHandler.async_threading::shutDown()) {
				async_threading::logDebug("com.symantec.premiumServices.asyncThreading.AsyncThreadManager", "shutting down...");
				return true;
			}
			
			async_threading::logDebug("com.symantec.premiumServices.asyncThreading.AsyncThreadManager", "unable to shut down...");
			return false;
		}
		
		/**
		 * retrieves the <code>AsyncThreadHandler</code> instance maintained by this singleton class.
		 *  
		 * @return AsyncThreadHandler 
		 * 
		 */
		public function get threadHandler():AsyncThreadHandler {
			return _threadHandler;
		}
		
		/**
		 * retrieves the <code>FPSObserverHandler</code> instance maintained by this singleton class.
		 *  
		 * @return FPSObserverHandler 
		 * 
		 */
		public function get observerHandler():FPSObserverHandler {
			return _observerHandler;
		}
		
	}
}

/**
 * this is class is not visible outside of the AsyncThreadingManager class, and is used 
 * as a key to enforce the constructor on the singleton instance.
 * 
 * @author ben rimbey -> ben_rimbey@symantec.com, ben.rimbey@gmail.com 
 * 
 */
class SingletonEnforcer {
	
	/**
	 * nothing really to document here - this class is never constructed. 
	 * 
	 */
	public function SingletonEnforcer():void {}
	
}