/**
 * Copyright (c) 2009 Symantec
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the 'Software'), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *  
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *  
 * THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
 
package com.symantec.premiumServices.asyncThreading.abstract {
	
	import com.symantec.premiumServices.asyncThreading.AsyncThreadingManager;
	import com.symantec.premiumServices.asyncThreading.events.AsyncThreadEvent;
	import com.symantec.premiumServices.asyncThreading.handlers.AsyncThreadHandler;
	import com.symantec.premiumServices.asyncThreading.interfaces.IAsyncThreadMessage;
	import com.symantec.premiumServices.asyncThreading.interfaces.IAsyncThreadResponder;
	import com.symantec.premiumServices.asyncThreading.namespaces.async_threading;
	
	import flash.events.EventDispatcher;
	import flash.utils.describeType;
	
	import mx.logging.Log;
	import mx.utils.UIDUtil;

	/**
	 * Dispatched when the thread is put to sleep
	 * 
	 * @eventType com.symantec.premiumServices.asyncThreading.events.AsyncThreadEvent
	*/
	[Event(name="sleep", type="com.symantec.premiumServices.asyncThreading.events.AsyncThreadEvent")]
	/**
	 * Dispatched when the thread is woken up
	 * 
	 * @eventType com.symantec.premiumServices.asyncThreading.events.AsyncThreadEvent
	*/
	[Event(name="wake", type="com.symantec.premiumServices.asyncThreading.events.AsyncThreadEvent")]
	/**
	 * Dispatched when the thread is started
	 * 
	 * @eventType com.symantec.premiumServices.asyncThreading.events.AsyncThreadEvent
	*/
	[Event(name="start", type="com.symantec.premiumServices.asyncThreading.events.AsyncThreadEvent")]
	/**
	 * Dispatched when the thread is terminated
	 * 
	 * @eventType com.symantec.premiumServices.asyncThreading.events.AsyncThreadEvent
	*/
	[Event(name="kill", type="com.symantec.premiumServices.asyncThreading.events.AsyncThreadEvent")]
	/**
	 * Dispatched when the thread has received a message
	 * 
	 * @eventType com.symantec.premiumServices.asyncThreading.events.AsyncThreadEvent
	 * @see IAsyncThreadMessage
	 */
	[Event(name="messageReceived", type="com.symantec.premiumServices.asyncThreading.events.AsyncThreadEvent")]
	/**
	 * This is the abstract base class used to create AsyncThread objects.
	 * 
	 * <p>Since this is an abstract class, it is absolutely essential that this class be subclassed by a class
	 * that implements either the IAsyncThreadResponder interface, or an interface that subclasses IAsyncThreadResponder, 
	 * such as IAsyncThreadScalableResponder. Failure to do this will result in a runtime exception when the extended class 
	 * is constructed.</p>
	 * 
	 * @example The following class extends <code>AbstractAsyncThread</code> and implements <code>IAsyncThreadScalableResponder</code>
	 * 
	 * <listing version="3.0">
	 * 
	 *			public class RunLevelNormalWorker extends AbstractAsyncThread implements IAsyncThreadScalableResponder {
	 *				
	 *				protected var iterationCount:int = 20000;
	 *				protected var binaryData:ByteArray = new ByteArray();
	 *				
	 *				public function RunLevelNormalWorker() {
	 *					super();
	 * 					super.priority = super.RUN_LEVEL_ABOVE_NORMAL;
	 *					super.start();
	 *				}
	 *				
	 * 				//enforced by IAsyncThreadResponder - invoked by Async_threadHandler
	 *				public function execute():void {
	 *					for (var i:int = 0; i < iterationCount; i++) {
	 *						binaryData.writeMultiByte("test", "unicodeFFFE");			
	 *					}			
	 *				} 
	 *			
	 * 				//enforced by IAsyncThreadScalableResponder - invoked by Async_threadHandler 
	 *				public function increaseWorkLoad():void {
	 *					iterationCount += 1000;
	 *				}
	 *				
	 * 				//enforced by IAsyncThreadScalableResponder - invoked by Async_threadHandler
	 *				public function decreaseWorkLoad():void {
	 *					iterationCount -= 1000;
	 *				}
	 *				
	 *			}
	 * 
	 * </listing>
	 * 
	 * @author ben rimbey -> ben_rimbey@symantec.com, ben.rimbey@gmail.com 
	 * @see IAsyncThreadResponder
	 * @see IAsyncThreadScalableResponder
	 */
	public class AbstractAsyncThread extends EventDispatcher {
		
		/**
		 * setting the thread to use this priority will make it execute on every frame. 
		 */
		public const RUN_LEVEL_REAL_TIME:int = 1;
		/**
		 * setting the thread to use this priority will make it execute on every 3rd frame. 
		 */
		public const RUN_LEVEL_HIGH:int = 3;
		/**
		 * setting the thread to use this priority will make it execute on every 6th frame. 
		 */
		public const RUN_LEVEL_ABOVE_NORMAL:int = 6;
		/**
		 * setting the thread to use this priority will make it execute on every 8th frame. 
		 */
		public const RUN_LEVEL_NORMAL:int = 8;
		/**
		 * setting the thread to use this priority will make it execute on every 12th frame. 
		 */
		public const RUN_LEVEL_BELOW_NORMAL:int = 12;
		/**
		 * setting the thread to use this priority will make it execute on every 24th frame. 
		 */
		public const RUN_LEVEL_LOW:int = 24;
		
		private var _sleeping:Boolean = false;
		private var _uid:String;
		private var _nthFrameToExecute:int = 8;
		private var _lastExecutionTime:int;
		private var _executionCount:int = 0;
		private var _messageInterests:Array = [];
		private var _messageQueue:Array = [];
		private var _distributedFrameSeed:int = 0;
		private var _threadHandler:AsyncThreadHandler = AsyncThreadingManager.getInstance().threadHandler;
		
		/**
		 * indicates whether or not the thread has been started yet. 
		 */
		protected var alive:Boolean = false;
		
		/**
		 * constructor - checks to make sure that the class instance implements the IAsyncThreadResponder interface, 
		 * or an interface extended from IAsyncThreadResponder, generates guid, then registers itself with the 
		 * AsyncThreadHandler. 
		 * @see AsyncThreadHandler
		 */
		public function AbstractAsyncThread() {
			if (this is IAsyncThreadResponder) {
				_uid = UIDUtil.createUID();
				
				inspectSelfForMessageResponders();
				_threadHandler.async_threading::registerThread(this);
				
			} else {
				throw new Error("The AbstractAsyncThread class must be extended by a subclass that implements the IAsyncThreadResponder interface.");
			}
		}
		
		/**
		 * sends a request to the AsyncThreadHandler to start this thread.
		 * 
		 * @see AsyncThreadHandler
		 */
		public function start():void {
			if (!alive) {
				_threadHandler.async_threading::startThread(this);
				alive = true;
				
				dispatchInternalEvent(AsyncThreadEvent.START);
			}
		}

		/**
		 * sends a request to the AsyncThreadHandler to put this thread to sleep. 
		 * 
		 * @see AsyncThreadHandler
		 */
		public function sleep():void {
			if (!_sleeping) {
				_sleeping = true;
				_threadHandler.async_threading::sleepThread(this);
				
				dispatchInternalEvent(AsyncThreadEvent.SLEEP);
			}			
		}
		
		/**
		 * sends a request to the AsyncThreadHandler to wake this thread up.
		 * 
		 * @see AsyncThreadHandler
		 */
		public function wake():void {
			if (_sleeping && alive) {
				_sleeping = false;
				_threadHandler.async_threading::wakeThread(this);
				
				dispatchInternalEvent(AsyncThreadEvent.WAKE);
			}
		}
		
		/**
		 * sends a request to the AsyncThreadHandler to kill this thread. 
		 * 
		 * @see AsyncThreadHandler
		 */
		public function kill():void {
			if (alive) {
				alive = false;
				_threadHandler.async_threading::killThread(this);
					
				dispatchInternalEvent(AsyncThreadEvent.KILL);
				delete(this);
			}
		}
		
		/**
		 * sets the current priority of this thread, where priority is an integer representing the nth frame for 
		 * this thread to execute on.
		 *  
		 * @param level
		 * 
		 */
		public function set priority(level:int):void {
			_nthFrameToExecute = level;
			AsyncThreadingManager.getInstance().async_threading::logDebug("com.symantec.premiumServices.asyncThreading.abstract.AbstractAsyncThread", this + " setting priority :: " + level + ", id :: " + this.id);
		}
		
		/**
		 * returns the current priority of this thread, where priority is an integer representing the nth frame 
		 * for this thread to execute on.
		 * 
		 * @return int
		 * 
		 */
		public function get priority():int {
			return _nthFrameToExecute;
		}
		
		/**
		 * returns the random seed used to offset the execution frame number for this thread. only used if 
		 * 'useDistributedFrameLoading' is enabled on AsyncThreadingManager.
		 * 
		 * @return int 
		 * @see AsyncThreadingManager
		 */
		public function get distributedFrameSeed():int {
			return _distributedFrameSeed;
		}
		
		/**
		 * whether or not this thread is currently idle.
		 * 
		 * @return Boolean
		 * 
		 */
		public function get sleeping():Boolean {
			return _sleeping;
		}
		
		/**
		 * the unique id for this thread.
		 * 
		 * @return String 
		 * 
		 */
		public function get id():String {
			return _uid;
		}
		
		/**
		 * the number of milliseconds it took to execute this thread during its last execution cycle.
		 * 
		 * @return int
		 * 
		 */
		public function get lastExecutionTime():int {
			return _lastExecutionTime;
		}
		
		/**
		 * the number of times this thread has executed.
		 * 
		 * @return int
		 * 
		 */
		public function get executionCount():int {
			return _executionCount;
		}
		
		/**
		 * a list of message types that this thread will respond to.
		 *  
		 * @return Array
		 * 
		 */
		public function get messageInterests():Array {
			return _messageInterests;
		}
		
		/**
		 * a list of pending messages for this thread.
		 * 
		 * @return Array 
		 * 
		 */
		protected function get messageQueue():Array {
			return _messageQueue;
		}
		
		/**
		 * iterates over pending messages in _messageQueue and invokes the appropriate methods to receive them.
		 * 
		 * @see IAsyncThreadMessage
		 */
		protected function handleMessageInbox():void {
			if (this.messageQueue.length > 0) {
				var vo:IAsyncThreadMessage;
				var responderMethodName:String;
				
				for (var i:int = 0; i < this.messageQueue.length; i++) {
					try {
						vo = this.messageQueue[i] as IAsyncThreadMessage;
						responderMethodName = "respondTo";
						responderMethodName += vo.name.substr(0, 1).toUpperCase();
						responderMethodName += vo.name.substr(1, vo.name.length);
						
						if (vo.body == null) {
							(this[responderMethodName] as Function).apply(this, [null]);
						} else {
							(this[responderMethodName] as Function).apply(this, [vo]);
						}
						
						var event:AsyncThreadEvent = new AsyncThreadEvent(AsyncThreadEvent.MESSAGE_RECEIVED, this as IAsyncThreadResponder);
							event.message = vo;
						super.dispatchEvent(event);
					} catch (e:Error) {
						Log.getLogger("com.symantec.premiumServices.asyncThreading.abstract.AbstractAsyncThread").debug(this + " unable to apply message to callback method :: " + responderMethodName);
					}
				}
			}
		}
		
		/**
		 * generate list of messages that this class is interested in by using describeType(this) to inspect the 
		 * method signature of this class instance. we only care about methods that begin with "respondTo".
		 * 
		 * @example this example responds to messages that are of type 'closeSocket'
		 * 
		 * <listing version="3.0">
		 * 
		 * 			//responds to messages with the type of "closeSocket"
		 *			public function respondToCloseSocket(msg:IAsyncThreadMessage):void {
		 * 				//do work here
		 * 			}
		 * 
		 * </listing>
		 * 
		 * @see describeType
		 * @see IAsyncThreadMessage
		 */
		protected function inspectSelfForMessageResponders():void {
			var map:XML = describeType(this);
			
			for each (var node:XML in map.method) {
				var name:String = node.@name;
				if (name.indexOf("respondTo") > -1) {
					name = name.substr(9, name.length);
					name = name.substr(0, 1).toLowerCase() + name.substr(1, name.length);
					
					_messageInterests.push(name);
				}
			}
		}
		
		/**
		 * sends a message from this thread instance to the rest of the threading framework.
		 * 
		 * @param name
		 * @param body
		 * @see AsyncThreadingManager
		 * @see IAsyncThreadMessage
		 */
		protected function sendThreadMessage(name:String, body:Object):void {
			AsyncThreadingManager.getInstance().sendMessageToThreads(name, body);
		}
		
		/**
		 * wrapper method to dispatch status eventing from this thread instance.
		 *  
		 * @param message
		 * 
		 */
		protected function dispatchInternalEvent(message:String):void {
			super.dispatchEvent(new AsyncThreadEvent(message, this as IAsyncThreadResponder));
		}
		
		/**
		 * invoked by the AsyncThreadManager to add messages to the pending queue for this thread instance.
		 *  
		 * @param message
		 * @see IAsyncThreadMessage
		 * @see AsyncThreadManager
		 */
		async_threading function addMessageToQueue(message:IAsyncThreadMessage):void {
			_messageQueue.push(message);
		}
		
		/**
		 * invoked by the AsyncThreadManager to execute this thread.
		 * <p>if this thread is not sleeping and it has been started, we first handle any pending messages. then
		 * we invoke the execute() method on the class extending this base class via the IAsyncThreadResponder 
		 * interface</p>
		 *  
		 * @see AsyncThreadManager
		 * @see handleMessageInbox
		 */
		async_threading function handleExecuteRequest():void {
			if (!this.sleeping && this.alive) {
				handleMessageInbox();
				
				(this as IAsyncThreadResponder).execute();
				_executionCount++;
				_messageQueue = [];
			}
		}
		
		/**
		 * invoked by the AsyncThreadManager to set the milliseconds it took to run this thread during its last 
		 * execution cycle.
		 *  
		 * @param val
		 * @see AsyncThreadManager
		 */
		async_threading function set lastExecutionTime(val:int):void {
			_lastExecutionTime = val;
		}
		
		/**
		 * invoked by the AsyncThreadManager to set the random frame seed for this thread instance.
		 *  
		 * @param val
		 * @see AsyncThreadManager
		 */
		async_threading function set distributedFrameSeed(val:int):void {
			_distributedFrameSeed = val;
		}
		
	}
}