/**
 * Copyright (c) 2009 Symantec
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the 'Software'), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *  
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *  
 * THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
 
 package com.symantec.premiumServices.asyncThreading.handlers {
	
	import com.symantec.premiumServices.asyncThreading.AsyncThreadingManager;
	import com.symantec.premiumServices.asyncThreading.abstract.AbstractAsyncThread;
	import com.symantec.premiumServices.asyncThreading.interfaces.IAsyncThreadMessage;
	import com.symantec.premiumServices.asyncThreading.interfaces.IAsyncThreadScalableResponder;
	import com.symantec.premiumServices.asyncThreading.namespaces.async_threading;
	import com.symantec.premiumServices.asyncThreading.utils.StopWatch;
	
	import mx.logging.Log;

	/**
	 * This class is instantiated and managed by the AsyncThreadingManager singleton. It manages all aspects
	 * of a thread's lifecycle, from registration to termination. It also routes incoming IAsyncThreadMessage 
	 * objects to the appropriate thread objects.
	 *  
	 * @author ben rimbey -> ben_rimbey@symantec.com, ben.rimbey@gmail.com 
	 * 
	 */
	public class AsyncThreadHandler {
		
		private var _messageBoxMap:Object = {};
		private var _registeredThreads:Object = {};
		private var _runningThreadIDs:Array = [];
		
		/**
		 * Constructor. This class is instantiated by the AsyncThreadingManager singleton. 
		 * 
		 */
		public function AsyncThreadHandler():void {
			AsyncThreadingManager.getInstance().async_threading::logDebug("com.symantec.premiumServices.asyncThreading.handlers.AsyncThreadHandler", "class instantiated...");
		}
		
		/**
		 * utility method used to retrieve an AbstractAsyncThread object by it's id.
		 * 
		 * @param id
		 * @return AbstractAsyncThread
		 * 
		 */
		public function getThread(id:String):AbstractAsyncThread {
			return registeredThreads[id] as AbstractAsyncThread;
		}
		
		/**
		 * this method is directly invoked by the FPSObserverHandler instance that is managed by the 
		 * AsyncThreadingManager singleton.
		 * 
		 * <p>every time the FPSObserverHandler catches an onEnterFrame event from its FPSObserver instance 
		 * that is monitoring the stage, it hits this method and passes in the current frame number that is being 
		 * tracked (ie: 1 through 24).</p>
		 * 
		 * <p>if AsyncThreadingManager.useDistributedFrameLoading is enabled, we distribute the way 
		 * in which we execute threads by adding a random seed value to the current frame number.</p>
		 * 
		 * <p>if this method succeeds in executing any threads, it pings the storeOverallExecutionTime() method 
		 * on the FPSObserverHandler instance in order to track execution times.</p>
		 *  
		 * @param frameCountForThisSecond
		 * 
		 */
		internal function executeThreads(frameCountForThisSecond:int):void {
			var executed:Boolean = false;
			var thread:AbstractAsyncThread = getThread(id);
			var executionTime:int = 0;
			var individualStopWatch:StopWatch = new StopWatch();
			var seededFrameCount:int = frameCountForThisSecond;
			
			for each (var id:String in runningThreadIDs) {
				thread = getThread(id);
				
				if (AsyncThreadingManager.useDistributedFrameLoading) {
					seededFrameCount = frameCountForThisSecond + thread.distributedFrameSeed;
				}
				
				if ((seededFrameCount % thread.priority) == 0) { 
					individualStopWatch.start();
					thread.async_threading::handleExecuteRequest();
					individualStopWatch.stop();
					
					evaluateThread(thread, individualStopWatch.lastRunTime);
					executed = true;
				}
			}
			
			if (executed) {
				AsyncThreadingManager.getInstance().observerHandler.storeOverallExecutionTime();
			}
		}
		
		/**
		 * evaluates a given thread. if the thread implements a descendent of the IAsyncThreadScalableResponder, 
		 * we look at several metrics to see if we need to increase or decrease its workload.
		 * 
		 * <p>invoked by the internal <code>executeThreads</code> method</p>
		 *  
		 * @param thread
		 * @param lastExecutionTime
		 * 
		 * @see IAsyncThreadScalableResponder
		 * @see executeThreads
		 */
		protected function evaluateThread(thread:AbstractAsyncThread, lastExecutionTime:int):void {
			thread.async_threading::lastExecutionTime = lastExecutionTime;
			
			if (thread is IAsyncThreadScalableResponder) {
				if (AsyncThreadingManager.getInstance().observerHandler.executedThreadFrameRate < AsyncThreadingManager.getInstance().moderateFrameRateThreshold) {
					decreaseThreadWorkLoad(thread);
					return;
				}
				
				if (AsyncThreadingManager.getInstance().observerHandler.currentFPS <= AsyncThreadingManager.getInstance().lowFrameRateThreshold) {
					decreaseThreadWorkLoad(thread);
					return;
				}
				
				if (thread.lastExecutionTime > AsyncThreadingManager.getInstance().medianExecutionTime && (AsyncThreadingManager.getInstance().observerHandler.currentFPS <= AsyncThreadingManager.getInstance().moderateFrameRateThreshold)) {
					decreaseThreadWorkLoad(thread);
					return;
				}
				
				if (thread.lastExecutionTime > AsyncThreadingManager.getInstance().maxExecutionTime) {
					decreaseThreadWorkLoad(thread);
					return;
				}
				
				if (AsyncThreadingManager.getInstance().observerHandler.currentFPS > AsyncThreadingManager.getInstance().moderateFrameRateThreshold) {
					increaseThreadWorkLoad(thread);
					return;
				}
			}
		}
		
		/**
		 * invoked by the <code>evaluteThread</code> method to decrease the workload on a given method.
		 * 
		 * @param thread
		 * @see IAsyncThreadScalableResponder
		 * @see evaluateThread
		 */
		protected function decreaseThreadWorkLoad(thread:AbstractAsyncThread):void {
			(thread as IAsyncThreadScalableResponder).decreaseWorkLoad();
			AsyncThreadingManager.getInstance().async_threading::logDebug("com.symantec.premiumServices.asyncThreading.handlers.AsyncThreadHandler", thread + " decreasing workload for thread, id :: " + thread.id);
		}
		
		/**
		 * invoked by the <code>evaluteThread</code> method to increase the workload on a given method.
		 * 
		 * @param thread
		 * @see IAsyncThreadScalableResponder
		 * @see evaluateThread
		 */
		protected function increaseThreadWorkLoad(thread:AbstractAsyncThread):void {
			(thread as IAsyncThreadScalableResponder).increaseWorkLoad();
			AsyncThreadingManager.getInstance().async_threading::logDebug("com.symantec.premiumServices.asyncThreading.handlers.AsyncThreadHandler", thread + " increasing workload for thread, id :: " + thread.id);
		}
		
		/**
		 * invoked by the <code>registerThread</code> method. parses a given thread's message interests 
		 * and adds those interests to existing mailboxes, if they exist. if not, we generate new mail boxes.
		 * 
		 * @param thread
		 * @see AbstractAsyncThread
		 * @see registerThread
		 */
		protected function initializeThreadMessageInterests(thread:AbstractAsyncThread):void {
			var name:String;
			for (var i:int = 0; i < thread.messageInterests.length; i++) {
				name = thread.messageInterests[i] as String;
				
				if (this.messageBoxMap[name] == null || this.messageBoxMap[name] == undefined) {
					this.messageBoxMap[name] = [];
				}
				
				(messageBoxMap[name] as Array).push(thread.id);
			}
		}
		
		/**
		 * utility method - generates random seed to distribute frame loading.
		 * 
		 * @param min
		 * @param max
		 * @return Number
		 * 
		 */
		private function getRandomFrameSeed(min:Number, max:Number):Number {
	        var scale:Number = max - min;
	        return Math.random() * scale + min;
	     }
		
		/**
		 * invoked by the <code>AsyncThreadingManager</code> singleton when a new <code>IAsyncThreadMessage</code> 
		 * is introduced to the system.
		 * 
		 * @param message
		 * 
		 */
		async_threading function handleInboundMessage(message:IAsyncThreadMessage):void {
			var targetThreadList:Array = messageBoxMap[message.name] as Array;
			var currentThread:AbstractAsyncThread;
			var currentId:String;
			
			if (targetThreadList != null) {
				for (var i:int = 0; i < targetThreadList.length; i++) {
					currentId = targetThreadList[i] as String;
					currentThread = getThread(currentId);
					
					currentThread.async_threading::addMessageToQueue(message);
				}
			}
		}
		
		/**
		 * invoked by <code>AbstractAsyncThread</code> instances. tries to register the thread in the 
		 * system. if it fails for some reason, we bounce out and return false.
		 * 
		 * @param thread
		 * @return Boolean
		 * @see AbstractAsyncThread
		 */
		async_threading function registerThread(thread:AbstractAsyncThread):Boolean {
			try {
				registeredThreads[thread.id] = thread;
				initializeThreadMessageInterests(thread);
				thread.async_threading::distributedFrameSeed = getRandomFrameSeed(0, 23);
				
				AsyncThreadingManager.getInstance().async_threading::logDebug("com.symantec.premiumServices.asyncThreading.handlers.AsyncThreadHandler", thread + " thread registered, id :: " + thread.id);
				
				return true;
			} catch (e:Error) {
				AsyncThreadingManager.getInstance().async_threading::logDebug("com.symantec.premiumServices.asyncThreading.handlers.AsyncThreadHandler", thread + " unable to register thread, id :: " + e.message);
			}
			
			return false;
		}
		
		/**
		 * invoked by the <code>AsyncThreadingManager</code> singleton. attempts to terminate 
		 * all registered threads.
		 * 
		 */
		async_threading function killAllThreads():void {
			for each (var thread:AbstractAsyncThread in registeredThreads) {
				async_threading::killThread(thread);
			}
		}
		
		/**
		 * invoked by <code>AbstractAsyncThread</code> instances. tries to start the thread in the 
		 * system. if it fails for some reason, we bounce out and return false.
		 * 
		 * @param thread
		 * @return Boolean
		 * @see AbstractAsyncThread
		 */
		async_threading function startThread(thread:AbstractAsyncThread):Boolean {
			try {
				runningThreadIDs.push(thread.id);
				AsyncThreadingManager.getInstance().async_threading::logDebug("com.symantec.premiumServices.asyncThreading.handlers.AsyncThreadHandler", thread + " starting thread, id :: " + thread.id);
			} catch (e:Error) {
				Log.getLogger("com.symantec.premiumServices.asyncThreading.handlers.AsyncThreadHandler").debug("UNABLE TO START THREAD :: " + e.getStackTrace());
				return false;
			}
			
			return true;
		}
		
		/**
		 * invoked by <code>AbstractAsyncThread</code> instances. tries to remove the thread from the 
		 * system. if it fails for some reason, we bounce out and return false.
		 * 
		 * @param thread
		 * @return Boolean
		 * @see AbstractAsyncThread
		 */
		async_threading function killThread(thread:AbstractAsyncThread):Boolean {
			try {	
				runningThreadIDs.splice(runningThreadIDs.indexOf(thread.id), 1);
				delete registeredThreads[thread.id];
				AsyncThreadingManager.getInstance().async_threading::logDebug("com.symantec.premiumServices.asyncThreading.handlers.AsyncThreadHandler", thread + " killing thread, id :: " + thread.id);
			} catch (e:Error) {
				Log.getLogger("com.symantec.premiumServices.asyncThreading.handlers.AsyncThreadHandler").debug("UNABLE TO KILL THREAD :: " + e.getStackTrace());
				return false;
			}
			
			return true;
		}
		
		/**
		 * invoked by <code>AbstractAsyncThread</code> instances. tries to put the thread to sleep. 
		 * if it fails for some reason, we bounce out and return false.
		 * 
		 * @param thread
		 * @return Boolean
		 * @see AbstractAsyncThread
		 */
		async_threading function sleepThread(thread:AbstractAsyncThread):Boolean {
			try {	
				runningThreadIDs.splice(runningThreadIDs.indexOf(thread.id), 1);
				AsyncThreadingManager.getInstance().async_threading::logDebug("com.symantec.premiumServices.asyncThreading.handlers.AsyncThreadHandler", thread + " sleeping thread, id :: " + thread.id);
			} catch (e:Error) {
				Log.getLogger("com.symantec.premiumServices.asyncThreading.handlers.AsyncThreadHandler").debug("UNABLE TO SLEEP THREAD :: " + e.getStackTrace());
				return false;
			}
			
			return true;
		}
		
		/**
		 * invoked by <code>AbstractAsyncThread</code> instances. tries to wake the thread and re-register it 
		 * in the system in a running state. if it fails for some reason, we bounce out and return false.
		 * 
		 * @param thread
		 * @return Boolean
		 * @see AbstractAsyncThread
		 */
		async_threading function wakeThread(thread:AbstractAsyncThread):Boolean {
			try {	
				runningThreadIDs.push(thread.id);
				AsyncThreadingManager.getInstance().async_threading::logDebug("com.symantec.premiumServices.asyncThreading.handlers.AsyncThreadHandler", thread + " waking thread, id :: " + thread.id);
			} catch (e:Error) {
				Log.getLogger("com.symantec.premiumServices.asyncThreading.handlers.AsyncThreadHandler").debug("UNABLE TO WAKE THREAD :: " + e.getStackTrace());
				return false;
			}
			
			return true;
		}
		
		/**
		 * invoked by the <code>AsyncThreadingManager</code> singleton. attempts to terminate 
		 * all registered threads, then reset all mappings.
		 * 
		 * @return Boolean
		 * @see AsyncThreadingManager
		 */
		async_threading function shutDown():Boolean {
			try {
				async_threading::killAllThreads();
				_messageBoxMap = {};
				_registeredThreads = {};
				_runningThreadIDs = [];	
				
				return true;
			} catch (e:Error) {
				Log.getLogger("com.symantec.premiumServices.asyncThreading.handlers.AsyncThreadHandler").debug("UNABLE TO SHUTDOWN AsyncThreadHandler :: " + e.getStackTrace());
			}
			
			return false;
		}
		
		
		/**
		 * retrieves a hashmap of all current mailboxes.
		 * 
		 * @return Object
		 * 
		 */
		public function get messageBoxMap():Object {
			return _messageBoxMap;
		}
		
		/**
		 * retrieves a hashmap of all registered threads.
		 * 
		 * @return Object
		 * 
		 */
		public function get registeredThreads():Object {
			return _registeredThreads;
		}
		
		/**
		 * retrieves an array containing the ids of all currently running threads.
		 * 
		 * @return Array
		 * 
		 */
		public function get runningThreadIDs():Array {
			return _runningThreadIDs;
		}
		
	}
}