/**
 * Copyright (c) 2009 Symantec
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the 'Software'), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *  
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *  
 * THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
 
package com.symantec.premiumServices.asyncThreading.handlers {
	import com.symantec.premiumServices.asyncThreading.AsyncThreadingManager;
	import com.symantec.premiumServices.asyncThreading.interfaces.IFPSObserver;
	import com.symantec.premiumServices.asyncThreading.namespaces.async_threading;
	import com.symantec.premiumServices.asyncThreading.observers.FPSObserver;
	import com.symantec.premiumServices.asyncThreading.observers.FPSReportQueue;
	
	import flash.display.DisplayObject;
	import flash.events.Event;
	
	import mx.core.Application;
	import mx.logging.Log;
	
	/**
	 * This class is instantiated and managed by the AsyncThreadingManager singleton. It primarily 
	 * monitors the stage framerate in order to gauge how much load we are putting on the system with 
	 * our thread instances. 
	 * 
	 * <p>additionally, it can monitor additional display objects's framerates. if any of them are low, 
	 * it messages that the entire system is low. this is to compensate for the way the AVM treats imported 
	 * assets from Flash that have different timelines; basically their framerate can get out of sync with 
	 * the framerate of the Flex parent application</p>
	 * 
	 * @author ben rimbey -> ben_rimbey@symantec.com, ben.rimbey@gmail.com 
	 * 
	 */
	public class FPSObserverHandler {
		
		private var _appRef:Application = Application.application as Application;
		private var _registeredUIMonitorList:Object = {};
		private var _stageFPSMonitor:FPSObserver;
		private var _executedFPSQueue:FPSReportQueue = new FPSReportQueue();
		private var _storeLastFrameTime:Boolean = false;
		
		/**
		 * Constructor. This class is instantiated by the AsyncThreadingManager singleton. 
		 * 
		 */
		public function FPSObserverHandler():void {
			AsyncThreadingManager.getInstance().async_threading::logDebug("com.symantec.premiumServices.asyncThreading.handlers.FPSObserverHandler", "class instantiated...");
		}
		
		/**
		 * callback method from the <code>FPSObserver</code> instance that is assigned to monitor 
		 * the stage reference. 
		 * 
		 */
		protected function handleEnterFrame():void {
			if (_storeLastFrameTime) {
				_executedFPSQueue.push(stageFPSMonitor.lastFPSReading);
				_storeLastFrameTime = false;
			}
			
			AsyncThreadingManager.getInstance().threadHandler.executeThreads(currentFrame);
		}
		
		/**
		 * invoked directly by the <code>AsyncThreadHandler</code> instance that is maintained by 
		 * the <code>AsyncThreadingManager</code> singleton.
		 * 
		 * <p>if the <code>AsyncThreadHandler</code> encounters any threads to execute in its exectution 
		 * schedule, it inokes this method to track the execution time. this provides a metric to 
		 * determine the overall impact the threading library is having on the application's performance.</p> 
		 * 
		 */
		internal function storeOverallExecutionTime():void {
			_storeLastFrameTime = true;
		}
		
		/**
		 * invoked by the <code>AsyncThreadingManager</code> singleton.
		 * 
		 * <p>instantiates the <code>FPSObserver</code> instance monitoring the stage.</p> 
		 * 
		 * @see AsyncThreadingManager
		 */
		async_threading function startUp():void {
			AsyncThreadingManager.getInstance().async_threading::logDebug("com.symantec.premiumServices.asyncThreading.handlers.FPSObserverHandler", "monitoring stage framerate...");
			_stageFPSMonitor = async_threading::registerUIComponent(_appRef.systemManager.stage as DisplayObject, handleEnterFrame);
		}
		
		/**
		 * invoked by the <code>AsyncThreadingManager</code> singleton. 
		 * 
		 * <p>registers a new <code>FPSObserver</code> instance to monitor the DisplayObject that is passed in 
		 * to the method. Also, if there is an external callback method supplied, we register that callback with 
		 * the newly instantiated <code>FPSObserver</code> instance.
		 * 
		 * @param uiRef
		 * @param enterFrame
		 * @return 
		 * @see AsyncThreadingManager
		 */
		async_threading function registerUIComponent(uiRef:DisplayObject, enterFrame:Function):IFPSObserver {
			var observer:FPSObserver = new FPSObserver();
				observer.start(uiRef, enterFrame);
				
			registeredUIMonitorList[uiRef.name] = observer;
			
			uiRef.addEventListener(Event.REMOVED_FROM_STAGE, handleRegisteredUIDeath);	
			
			return observer;
		}
		
		/**
		 * invoked either by the <code>AsyncThreadingManager</code> singleton, or by the <code>handleRegisteredUIDeath</code> method. 
		 * 
		 * <p>unregisters the <code>FPSObserver</code> instance which is registered to monitor the DisplayObject 
		 * whose name is passed into this method.
		 * 
		 * @param uiRefName
		 * @see AsyncThreadingManager
		 */
		async_threading function unregisterUIComponent(uiRefName:String):void {
			var observer:FPSObserver = (registeredUIMonitorList[uiRefName] as FPSObserver);
				observer.kill();
			
			delete registeredUIMonitorList[uiRefName];
		}
		
		/**
		 * invoked by the <code>AsyncThreadingManager</code> singleton. 
		 * 
		 * <p>attempts to kill the <code>FPSMonitor</code> which is monitoring the stage reference, as well 
		 * as attempting to kill every <code>FPSMonitor</code> which is monitoring additional assets. 
		 * additionally, we reset the values of all internally tracked objects.</p>
		 *  
		 * @return 
		 * @see AsyncThreadingManager
		 */
		async_threading function shutDown():Boolean {
			try {
				if (stageFPSMonitor != null) {
					stageFPSMonitor.kill();
				}
				
				for each (var observer:IFPSObserver in registeredUIMonitorList) {
					observer.kill();
				}
				
				_registeredUIMonitorList = {};
				_stageFPSMonitor = null;
				_executedFPSQueue = new FPSReportQueue;
				_storeLastFrameTime = false;
				
				return true;
			} catch (e:Error) {
				Log.getLogger("com.symantec.premiumServices.asyncThreading.handlers.FPSObserverHandler").debug("UNABLE TO SHUTDOWN FPSObserverHandler :: " + e.getStackTrace());
			}
			
			return false;
		}
		
		/**
		 * event callback triggered by the death of a DisplayObject whose framerate we are monitoring.
		 *  
		 * @param evn
		 * 
		 */
		protected function handleRegisteredUIDeath(evn:Event):void {
			AsyncThreadingManager.getInstance().async_threading::logDebug("com.symantec.premiumServices.asyncThreading.handlers.FPSObserverHandler", "caught ui component death, removing associated fps observer...");
			async_threading::unregisterUIComponent((evn.target as DisplayObject).name);
		}
		
		/**
		 * returns the integer value of the current frame in this second for the stage reference.
		 *  
		 * @return int
		 * 
		 */
		public function get currentFrame():int {
			return stageFPSMonitor.currentFrame;
		}
		
		/**
		 * returns the lowest value returned from all running <code>FPSObserver</code> instances. the 
		 * idea is that if one part of the system is slowing down, we message that the system as a whole 
		 * is slowing down.
		 *  
		 * @return int
		 * 
		 */
		public function get currentFPS():int {
			var base:int = stageFPSMonitor.lastFPSReading;
			
			for each (var observer:FPSObserver in registeredUIMonitorList) {
				if (observer.lastFPSReading < base) {
					base = observer.lastFPSReading;
				}
			}
			
			return base;
		}
		
		/**
		 * returns an averaged length of time in milliseconds for executing all threads during a given thread 
		 * execution cycle within the <code>AsyncThreadHandler</code> instance.
		 * 
		 * @return int 
		 * 
		 */
		public function get executedThreadFrameRate():int {
			return _executedFPSQueue.averageFPS;
		}
		
		/**
		 * retrieves a hashmap containing all of the currently registered <code>FPSObserver</code> instances.
		 * 
		 * @return Object 
		 * 
		 */
		public function get registeredUIMonitorList():Object {
			return _registeredUIMonitorList;
		}
		
		/**
		 * retrieves the <code>FPSObserver</code> instance assigned to monitor the stage. 
		 * 
		 * @return FPSObserver 
		 * 
		 */
		public function get stageFPSMonitor():FPSObserver {
			return _stageFPSMonitor;
		}
		
	}
}