/**
 * Copyright (c) 2009 Symantec
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the 'Software'), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *  
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *  
 * THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package com.symantec.premiumServices.asyncThreading.observers {
	import com.symantec.premiumServices.asyncThreading.interfaces.IFPSObserver;
	import com.symantec.premiumServices.asyncThreading.utils.StopWatch;
	
	import flash.display.DisplayObject;
	import flash.events.Event;
	
	/**
	 * this class observess a given display object and tracks its 
	 * <code>Event.ENTER_FRAME</code> events in an attempt to gather metrics about 
	 * its framerate.
	 * 
	 * @author ben rimbey -> ben_rimbey@symantec.com, ben.rimbey@gmail.com 
	 * 
	 */
	public class FPSObserver implements IFPSObserver {
		
		/**
		 * this method allows this class to notify outside classes of a <code>Event.ENTER_FRAME</code> 
		 * event. why don't we use events to do this? using callbacks locks the thread and guarantees the 
		 * order in which things are executed, which is important for tracking execution timing. 
		 */
		protected var enterFrameCallBack:Function; 
		/**
		 * this queue tracks the fps values retrieved by this FPSObserver instance.
		 */
		protected var fpsQueue:FPSReportQueue = new FPSReportQueue();
		/**
		 * utility object used to track how long it takes to render a frame. 
		 */
		protected var stopWatch:StopWatch = new StopWatch();
		
		private var _frameRate:int;
		private var _currentFrame:int = 0;		
		private var _uiRef:DisplayObject;
		
		/**
		 * called externally - typically by the <code>FPSObserverHandler</code> instance maintained 
		 * by the <code>AsyncThreadingManager</code> singleton.
		 * 
		 * <p>uiRef is the object which we want to observe for framerate. enterFrameCallBack is an 
		 * optional callback method parameter.</p>
		 *  
		 * @param uiRef
		 * @param enterFrameCallBack
		 * 
		 */
		public function start(uiRef:DisplayObject, enterFrameCallBack:Function=null):void {
			_uiRef = uiRef;
			this.enterFrameCallBack = enterFrameCallBack;
			
			_frameRate = uiRef.stage.frameRate;
			uiRef.addEventListener(Event.ENTER_FRAME, handleRenderFrame);
		}
		
		/**
		 * removes the <code>Event.ENTER_FRAME</code> event listener from the DisplayObject which 
		 * is being tracked. if for some reason that fails, we return <code>false</code>.
		 * 
		 * @return Boolean 
		 * 
		 */
		public function kill():Boolean {
			try {
				if (uiRef != null) {
					uiRef.removeEventListener(Event.ENTER_FRAME, handleRenderFrame);
				}
				return true;
			} catch (e:Error) {
				
			}
			return false;
		}
		
		/**
		 * callback for the <code>Event.ENTER_FRAME</code> event which is dispatched by the 
		 * DisplayObject being tracked.
		 * 
		 * <p>first we invoke <code>handleEndOfLastFrame</code>, then we invoke 
		 * <code>handleStartOfCurrentFrame</code>.
		 *  
		 * @param evn
		 * @see handleEndOfLastFrame
		 * @see handleStartOfCurrentFrame
		 */
		protected function handleRenderFrame(evn:Event):void {
			handleEndOfLastFrame();
			handleStartOfCurrentFrame();			
		}
		
		/**
		 * invoked by the <code>handleRenderFrame</code> method. here we stop the <code>StopWatch</code> 
		 * instance, and store its calculated FPS value into our <code>FPSReportQueue</code>. 
		 * 
		 * @see StopWatch
		 * @see FPSReportQueue
		 * @see StopWatch.getCalculatedFPS
		 */
		protected function handleEndOfLastFrame():void {
			stopWatch.stop();
			
			if (currentFrame > 0) {
				fpsQueue.push(stopWatch.getCalculatedFPS());
			}
		}
		
		/**
		 * invoked by the <code>handleRenderFrame</code> method. here we restart the <code>StopWatch</code> 
		 * instance. we also increment the value of the currentFrame that is being tracked.
		 * 
		 * <p>if a callback method was passed into the <code>start</code> method, we will 
		 * attempt to invoke it.</p>
		 * 
		 */
		protected function handleStartOfCurrentFrame():void {
			stopWatch.start();
			
			_currentFrame++;
			if (currentFrame > _frameRate) {
				_currentFrame = 1;
			}
			
			if (enterFrameCallBack != null) {
				enterFrameCallBack.apply(this, []);
			}
		}
		
		/**
		 * returns an average of the values stored in the <code>FPSReportQueue</code> instance.
		 * 
		 * @return int
		 * 
		 */
		public function get fps():int {
			return fpsQueue.averageFPS;
		}
		
		/**
		 * retrieves the last value that was pushed into the <code>FPSReportQueue</code> instance.
		 * 
		 * @return int 
		 * 
		 */
		public function get lastFPSReading():int {
			return fpsQueue[fpsQueue.length - 1] as int;
		}
		
		/**
		 * retrieves the value of the current frame, ie: 1 - 24.
		 * 
		 * @return int 
		 * 
		 */
		public function get currentFrame():int {
			return _currentFrame;
		}
		
		/**
		 * retrieves the DisplayObject being observed by this <code>FPSObserver</code> instance.
		 * 
		 * @return DisplayObject 
		 * 
		 */
		public function get uiRef():DisplayObject {
			return _uiRef;
		}

	}
}